
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include "ntcli.h"

static char * progname = "cat";

static int unbuffered_output = 0;
static int silent_mode = 0;
static int visible_mode = 0;
static int tabs_mode = 0;
static int eol_mode = 0;
static int text_mode = 0;

static int
CatOneFile(const char * fn)
{
   FILE * infile;
   if ((fn[0] == '-') && (fn[1] == '\0')) {
      infile = stdin;
   } else {
      infile = fopen(fn, "r");
   }
   if (infile != NULL) {
      _setmode(_fileno(infile), text_mode ? _O_TEXT : _O_BINARY);
      if (visible_mode) {
         /* If we're gonna screw around with translating chars and such,
         ** might as well just do the copy with getc and putc...
         */
         int c;
         while ((c = fgetc(infile)) != EOF) {
            if (isprint(c)) {
               /* We do not need to look any farther, this is a printable
               ** character, so we go ahead and print it...
               */
               fputc(c, stdout);
            } else {
               /* There are questions to be answered before we can print this
               ** character, so find the answers...
               */
               if (c == '\n') {
                  if (eol_mode) {
                     fputc('$', stdout);
                  }
                  fputc(c, stdout);
               } else if ((! tabs_mode) && ((c == '\t') || (c == '\f'))) {
                  fputc(c, stdout);
               } else {
                  if (c & 0x80) {
                     fputc('M', stdout);
                     fputc('-', stdout);
                     c &= (~ 0x80);
                  }
                  if ((c >= 0) && (c < 037)) {
                     fputc('^', stdout);
                     c += '@';
                  } else if (c == 0177) {
                     fputc('^', stdout);
                     c = '?';
                  }
                  fputc(c, stdout);
               }
            }
         }
      } else {
         /* OK, this is mega clean just shove bits out the door mode. Here
         ** I can grab big hunks and write them out fast as possible...
         ** (Should think about memory mapping input files, but I'll leave
         ** that alone for now).
         */
         char buf[32768];
         size_t incount;
         while ((incount = fread(buf, 1, sizeof(buf), infile)) > 0) {
            fwrite(buf, 1, incount, stdout);
         }
      }
      if (infile != stdin) {
         fclose(infile);
      }
   } else {
      if (! silent_mode) {
         fprintf(stderr, "%s: can't open %s\n", progname, fn);
      }
      return 1;
   }
   return 0;
}

int
main(int argc, char ** argv)
{
   NTCLIOptinfo oi;
   char * onearg;
   int errcount = 0;

   /* I'll be trying to wildcard expand things which might be on removable
   // devices, so tell the system to take a hike when it comes to displaying
   // error pop-up window's and just return an error code from the system
   // service instead.
   */
   SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOOPENFILEERRORBOX);

   oi = NTCLIInitOptinfo(argc, argv, "usvteTB");
   progname = NTCLIGetProgname(oi);
   for ( ; ; ) {
      int opt = NTCLIGetNextOption(oi);
      if (opt == -1) {
         char * msg = NTCLIErrorString(oi);
         if (msg != NULL) {
            fprintf(stderr, "%s\n", msg);
            exit(2);
         } else {
            break;
         }
      } else {
         switch(opt) {
         case 'u':      /* make output unbuffered */
            unbuffered_output = 1;
            break;
         case 's':      /* ignore non-existant files */
            silent_mode = 1;
            break;
         case 'v':      /* print non-printings chars as escape codes */
            visible_mode = 1;
            text_mode = 1;
            break;
         case 't':      /* print tabs and form feeds as ^I and ^L */
            tabs_mode = 1;
            visible_mode = 1;
            text_mode = 1;
            break;
         case 'e':      /* print $ just before newline */
            eol_mode = 1;
            visible_mode = 1;
            text_mode = 1;
            break;
         case 'T':      /* read and write files in text mode */
            text_mode = 1;
            break;
         case 'B':      /* read and write files in binary mode */
            text_mode = 0;
            break;
         }
      }
   }

   /* Setup stdout modes as requested
    */
   if (unbuffered_output) {
      setvbuf(stdout, NULL, _IONBF, 0);
   }
   _setmode(_fileno(stdout), text_mode ? _O_TEXT : _O_BINARY);

   /* Now expand the arguments and process the files.
    */
   NTCLIExpandWildcards(oi);
   onearg = NTCLIGetNextPositionalArg(oi);
   if (onearg == NULL) {
      /* No arguments - process stdin as input file */
      errcount += CatOneFile("-");
   } else {
      /* Process onearg followed by each additional arg as input files */
      while (onearg != NULL) {
         errcount += CatOneFile(onearg);
         onearg = NTCLIGetNextPositionalArg(oi);
      }
   }
   return ((errcount > 0) ? 2 : 0);
}
