
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include "ntcli.h"

static char * progname = "cmp";

static int silent_mode = 0;
static int long_mode = 0;
static int text_mode = 0;

static FILE *
OpenArg(const char * fn)
{
   FILE * rval;
   if ((fn[0] == '-') && (fn[1] == '\0')) {
      rval = stdin;
   } else {
      rval = fopen(fn, "r");
      if (rval == NULL) {
         fprintf(stderr, "%s: Can't open %s\n", progname, fn);
         exit(2);
      }
   }
   _setmode(_fileno(rval), text_mode ? _O_TEXT : _O_BINARY);
   return rval;
}

int
main(int argc, char ** argv)
{
   NTCLIOptinfo oi;
   char * onearg = NULL;
   char * arg1;
   char * arg2 = NULL;
   FILE * file1;
   FILE * file2;
   int c1, c2;
   size_t count = 0;
   int rcode = 0;

   /* I'll be trying to wildcard expand things which might be on removable
   // devices, so tell the system to take a hike when it comes to displaying
   // error pop-up window's and just return an error code from the system
   // service instead.
   */
   SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOOPENFILEERRORBOX);

   oi = NTCLIInitOptinfo(argc, argv, "lsTB");
   progname = NTCLIGetProgname(oi);
   for ( ; ; ) {
      int opt = NTCLIGetNextOption(oi);
      if (opt == -1) {
         char * msg = NTCLIErrorString(oi);
         if (msg != NULL) {
            fprintf(stderr, "%s\n", msg);
            exit(2);
         } else {
            break;
         }
      } else {
         switch(opt) {
         case 'l':      /* long listing of diffs */
            long_mode = 1;
            break;
         case 's':      /* silent - just set exit code */
            silent_mode = 1;
            break;
         case 'T':      /* read files in text mode */
            text_mode = 1;
            break;
         case 'B':      /* read files in binary mode */
            text_mode = 0;
            break;
         }
      }
   }

   /* Now expand the arguments and process the files.
    */
   NTCLIExpandWildcards(oi);
   arg1 = NTCLIGetNextPositionalArg(oi);
   if (arg1 != NULL) {
      arg2 = NTCLIGetNextPositionalArg(oi);
      if (arg2 != NULL) {
         onearg = NTCLIGetNextPositionalArg(oi);
         if (onearg != NULL) {
            fprintf(stderr, "%s: Too many arguments.\n", progname);
            exit(2);
         }
      }
   }
   if ((arg1 == NULL) || (arg2 == NULL)) {
      fprintf(stderr, "%s: Missing arguments.\n", progname);
      exit(2);
   }
   file1 = OpenArg(arg1);
   file2 = OpenArg(arg2);
   if ((file1 == stdin) && (file2 == stdin)) {
      fprintf(stderr, "%s: Only one arg should be stdin.\n", progname);
      exit(2);
   }
   for ( ; ; ) {
      c1 = fgetc(file1);
      c2 = fgetc(file2);
      if ((c1 == EOF) || (c2 == EOF)) {
         if (c1 != EOF) {
            fprintf(stdout, "%s: EOF in %s at byte %ld.\n",
                    progname, arg2, count);
            return 1;
         } else if (c2 != EOF) {
            fprintf(stdout, "%s: EOF in %s at byte %ld.\n",
                    progname, arg1, count);
            return 1;
         }
         return 0;
      }
      ++count;
      if (c1 != c2) {
         fprintf(stdout, "%s: %s and %s differ at byte %ld.\n",
                 progname, arg1, arg2, count);
         rcode = 1;
         if (! long_mode) {
            break;
         }
      }
   }
   return rcode;
}
