
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <windows.h>
#include "ntcli.h"

/****************************************************************** Features

This is an NT-aware version of ls. It acts a lot like the U**x ls, but deals
(I hope) intelligently with the NT filesystem. It will accept names in
either the forward or backslash format, and it can print the results in
either format. It can use the existing case of names, or it can convert
names to all upper or lower case on output (which also effects the way it
sorts the names). It can also print the short name MSDOS complaint aliases
for long NT file names (and with the -l and -o options, shows both names).

The program also incorporates a U**x-like argv expander so that much more
sophisticated wild-card expansion can be used in the arguments to ls.
Things like [cde]:\*\lib will expand to match all 2nd level subdirectories
named 'lib' on any of the c:, d: or e: filesystems. And the wildcard
expansion is case-insensitive to reflect the case insensitivity of the
filesystems.

This is a genuine Win32 program, using real NT API calls for digging up info
(not fake U**x stat() and readdir() emulations). OK, so its not much of a
feature, but it is the source of one of the few time saving features of the
program since the FindFirstFile interface returns enough information from
the directory that I don't have to do an additional stat-like call to get
size/time/etc.

*/

/************************************************************** Bug Features

I'm not sure if these count as bugs or features, so they are listed in their
own section :-).

Because NT file security is radically different from U**x security, it makes
no sense to exactly emulate the ls -l format of U**x, so the -l format has
more of an NT flavor. (Of course, you could argue that exact emulation is
useful, especially if you want to interpret the output of the program with
another program that expects U**x ls info, but you can't have everything).

Due to the awful penalties required to lookup owner and group info, you
have to explicitly turn on the display via the -G and -O options (which
is kind of backwards).

*/

/********************************************************************** Bugs

This is not the fastest code in the world. I suspect it would benefit a lot
from having its own custom memory management routines for tracking lists of
file names, and it could also be tweaked to avoid many of the uses of the
str* routines by keeping rack of string lengths. I just wanted to get an ls
that provided the features I wanted, not necessarily extract the last cycle
of performance (although as a native NT app, it sure runs faster than the
POSIX ls which has to load the whole posix subsystem in order to execute
:-).

This isn't a bug in this program as much as it is a bug in NT, but it
takes about 10 system service calls for *each* file to dig up the printable
info for user and group names, so the -G and -O options can be really slow...

The code is not really compatible with unicode. I use TCHAR in a few places,
but I didn't make any real attempt to get it unicode ready. Perhaps I will
make the changes someday...

*/

/*********************************************************** Data Structures */

/* OneFile holds the directory info describing a single file and can
// be part of a list of files.
*/
struct one_file_struct;
typedef struct one_file_struct {
   struct one_file_struct * next;       /* pointer to next in list */
   DWORD    FileAttributes;             /* attributes from directory */
   FILETIME CreationTime;
   FILETIME LastAccessTime;
   FILETIME LastWriteTime;
   DWORD    FileSizeHigh;
   DWORD    FileSizeLow;
   int      ShortNameIndex;             /* index in FileName of short name */
   TCHAR    FileName[1];                /* var sized array holding names */
} OneFile;

/* FileList holds a list of OneFile structs. This list can represent the
// contents of a directory, or can be used to hold the initial argument list
// for processing.
*/
struct file_list_struct;
typedef struct file_list_struct {
   OneFile *  head;                     /* head of list */
   OneFile ** tail;                     /* points to end ptr of list */
   DWORD      count;                    /* holds count of files in list */
   OneFile *  dirname;                  /* name of this directory (if named) */
} FileList;

/* OneDir records a directory name waiting to be processed and indicates if
// the name should be printed or not when processed.
*/
struct one_dir_struct;
typedef struct one_dir_struct {
   struct one_dir_struct * next;
   DWORD print_ndx;
   TCHAR dirname[1];
} OneDir;

/* DirList records a list of directories waiting to be processed.
*/
struct dir_list_struct;
typedef struct dir_list_struct {
   OneDir * head;
} DirList;

/* IgnoreList keeps a list of pointers to ignore patterns.
*/
struct ignore_list_struct;
typedef struct ignore_list_struct {
   struct ignore_list_struct * next;
   char * pattern;
} IgnoreList;

/********************************************************** Global Variables */

DirList worklist;

char * progname = "ls";

char * screen_width_string = NULL;

DWORD screen_width = 80;

BOOL comma_separate = FALSE;
BOOL convert_to_lower = FALSE;
BOOL do_not_sort = FALSE;
BOOL escape_unprintable = FALSE;
BOOL flag_dirs = FALSE;
BOOL flag_executables = FALSE;
BOOL list_across = FALSE;
BOOL list_by_columns = FALSE;
BOOL long_format = FALSE;
BOOL question_unprintable = FALSE;
BOOL quote_names = FALSE;
BOOL recursive_list = FALSE;
BOOL reverse_sort = FALSE;
BOOL show_both_names = FALSE;
BOOL show_dir_contents = TRUE;
BOOL show_dot_dot = FALSE;
BOOL show_group = FALSE;
BOOL show_hidden = FALSE;
BOOL show_owner = FALSE;
BOOL show_short_names = FALSE;
BOOL sort_by_extension = FALSE;
BOOL sort_by_size = FALSE;
BOOL upcase_names = FALSE;
BOOL use_access_time = FALSE;
BOOL use_backslash_sep = FALSE;
BOOL use_create_time = FALSE;
BOOL use_modtime = FALSE;

char * print_buf = NULL;
DWORD print_buf_len = 0;

TCHAR * current_directory_name = ".";

IgnoreList * ignore_list = NULL;

/********************************************** Routines to Manipulate Lists */

/* AllocFileList - create and initialize a new FileList.
*/
FileList *
AllocFileList()
{
   FileList * fl = (FileList *)NTCLImalloc(sizeof(FileList));
   fl->head = NULL;
   fl->tail = &fl->head;
   fl->count = 0;
   fl->dirname = NULL;
   return fl;
}

/* FreeFileList - reclaim storage used by FileList and the OneFile objects
// pointed at by the list.
*/
void
FreeFileList(FileList * fl)
{
   while (fl->head != NULL) {
      OneFile * nof = fl->head->next;
      NTCLIfree((void *)fl->head);
      fl->head = nof;
   }
   if (fl->dirname != NULL) {
      NTCLIfree((void *)fl->dirname);
   }
   NTCLIfree((void *)fl);
}

/* ReverseFileList - reverse the entries in a FileList
*/
void
ReverseFileList(FileList * fl)
{
   OneFile * ofp;
   if (fl->count > 1) {
      ofp = fl->head;
      fl->head = NULL;
      fl->tail = &ofp->next;
      while (ofp != NULL) {
         OneFile * nfp = ofp->next;
         ofp->next = fl->head;
         fl->head = ofp;
         ofp = nfp;
      }
   }
}

/* SortFileList sorts the list of files (given a sort routine to pass to
// qsort).
*/
void
SortFileList(
   FileList * fl,
   int (__cdecl * sortfunc)(const void *, const void *))
{
   OneFile * * ofl;
   OneFile * ofp;
   DWORD i;

   if (fl->count <= 1) {
      /* No need to sort zero or one element lists */
      return;
   }

   /* Allocate array to store pointers in, sort the array, copy back to the
   // list in sorted order, free the array.
   */

   ofl = (OneFile * *)NTCLImalloc(sizeof(OneFile *) * fl->count);
   i = 0;
   ofp = fl->head;
   while (ofp != NULL) {
      ofl[i++] = ofp;
      ofp = ofp->next;
   }
   qsort((void *)ofl, fl->count, sizeof(OneFile *), sortfunc);
   fl->head = NULL;
   fl->tail = &fl->head;
   for (i = 0; i < fl->count; ++i) {
      *(fl->tail) = ofl[i];
      ofl[i]->next = NULL;
      fl->tail = &(ofl[i]->next);
   }
   NTCLIfree((void *)ofl);
}

OneDir *
PopWorkList()
{
   OneDir * odp = worklist.head;
   if (odp != NULL) {
      worklist.head = odp->next;
      odp->next = NULL;
   }
   return odp;
}

/*************************************************** Routines to Build Lists */

/* FindOneFile just does a FindFirstFile to get the data for one name, then
// immediately closes the handle that creates (because we only want data for
// one name). This routine also incorporates the kludge for asking for info
// on root directories. NT (conveniently) doesn't allow FindFirstFile to
// work on strings like "\\" or "C:\\" (don't ask why, I don't know), so if
// the first test fails, I check and see if the file is a directory I can
// read, and if so, I build up an imaginary set of find data for it (getting
// things like create time by figuring that the oldest file in the directory
// is the create time of the directory, etc).
*/
BOOL
FindOneFile(TCHAR * name, WIN32_FIND_DATA * fdp)
{
   HANDLE h = FindFirstFile(name, fdp);
   if (h == INVALID_HANDLE_VALUE) {
      /* Well, it didn't work the first time, stick a "\\*.*" on the end and
      // see if it really is a directory, but I'm just not allowed to ask
      // about the name itself...
      */
      DWORD len = strlen(name);
      TCHAR * newname = NTCLImalloc(len + 6);
      WIN32_FIND_DATA fd;

      strcpy(newname, name);
      if (newname[len-1] != '\\') {
         newname[len++] = '\\';
      }
      strcpy(&newname[len], "*.*");
      memset((void *)fdp, 0, sizeof(WIN32_FIND_DATA));
      h = FindFirstFile(newname, &fd);
      if (h != INVALID_HANDLE_VALUE) {
         /* HA! It is a directory after all, set the attributes so it looks
         // like one, and then compute the create time as the oldest time of
         // any file, the last write time as the most recent create time of
         // any file, and the last access time as the newest time of any
         // file.
         */
         strcpy(fdp->cFileName, name);
         fdp->dwFileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
         for ( ; ; ) {
            if ((fd.ftLastAccessTime.dwLowDateTime == 0) &&
                (fd.ftLastAccessTime.dwHighDateTime == 0)) {
               fd.ftLastAccessTime = fd.ftLastWriteTime;
            }
            if ((fd.ftCreationTime.dwLowDateTime == 0) &&
                (fd.ftCreationTime.dwHighDateTime == 0)) {
               fd.ftCreationTime = fd.ftLastAccessTime;
            }
            if (((fdp->ftLastAccessTime.dwLowDateTime == 0) &&
                 (fdp->ftLastAccessTime.dwHighDateTime == 0)) ||
                (CompareFileTime(&fdp->ftLastAccessTime,
                                 &fd.ftLastAccessTime) < 0)) {
               fdp->ftLastAccessTime = fd.ftLastAccessTime;
            }
            if (((fdp->ftLastWriteTime.dwLowDateTime == 0) &&
                 (fdp->ftLastWriteTime.dwHighDateTime == 0)) ||
                (CompareFileTime(&fdp->ftLastWriteTime,
                                 &fd.ftLastWriteTime) < 0)) {
               fdp->ftLastWriteTime = fd.ftLastWriteTime;
            }
            if (((fdp->ftCreationTime.dwLowDateTime == 0) &&
                 (fdp->ftCreationTime.dwHighDateTime == 0)) ||
                (CompareFileTime(&fdp->ftCreationTime,
                                 &fd.ftCreationTime) > 0)) {
               fdp->ftCreationTime = fd.ftCreationTime;
            }
            if (! FindNextFile(h, &fd)) {
               break;
            }
         }
         FindClose(h);
         return TRUE;
      }
   } else {
      FindClose(h);
      return TRUE;
   }
   return FALSE;
}

/* PutStringOnList takes a file name string, looks up the directory info for
// that file, and makes a list entry for the file recording the full name
// and the additional directory info of interest. It is used primarily for
// processing the initial argument list.
*/
void
PutStringOnList(FileList * fl, TCHAR * name)
{
   WIN32_FIND_DATA fd;

   if (! FindOneFile(name, &fd)) {
      fprintf(stderr, "File not found: %s\n", name);
   } else {
      DWORD lnl = strlen(name);
      DWORD snl = strlen(fd.cAlternateFileName);
      DWORD tlen = lnl + snl + 1;
      OneFile * ofp;

      if (snl > 0) {
         ++tlen;
      }
      ofp = (OneFile *)NTCLImalloc(sizeof(OneFile) + tlen - 1);
      ofp->next = NULL;
      ofp->FileAttributes = fd.dwFileAttributes;
      ofp->CreationTime = fd.ftCreationTime;
      ofp->LastAccessTime = fd.ftLastAccessTime;
      ofp->LastWriteTime = fd.ftLastWriteTime;
      ofp->FileSizeHigh = fd.nFileSizeHigh;
      ofp->FileSizeLow = fd.nFileSizeLow;
      strcpy(ofp->FileName, name);
      if (snl > 0) {
         ofp->ShortNameIndex = lnl + 1;
         strcpy(&ofp->FileName[ofp->ShortNameIndex], fd.cAlternateFileName);
      } else {
         ofp->ShortNameIndex = lnl;
      }
      *fl->tail = ofp;
      fl->tail = &ofp->next;
      fl->count += 1;
   }
}

/* PutFindDataOnList is used to accumulate each new directory entry when
// reading a directory via FindFirstFile/FindNextFile
*/
void
PutFindDataOnList(FileList * fl, WIN32_FIND_DATA * fdp)
{
   DWORD lnl = strlen(fdp->cFileName);
   DWORD snl = strlen(fdp->cAlternateFileName);
   DWORD tlen = lnl + snl + 1;
   OneFile * ofp;

   if (snl > 0) {
      ++tlen;
   }
   ofp = (OneFile *)NTCLImalloc(sizeof(OneFile) + tlen - 1);
   ofp->next = NULL;
   ofp->FileAttributes = fdp->dwFileAttributes;
   ofp->CreationTime = fdp->ftCreationTime;
   ofp->LastAccessTime = fdp->ftLastAccessTime;
   ofp->LastWriteTime = fdp->ftLastWriteTime;
   ofp->FileSizeHigh = fdp->nFileSizeHigh;
   ofp->FileSizeLow = fdp->nFileSizeLow;
   strcpy(ofp->FileName, fdp->cFileName);
   if (snl > 0) {
      ofp->ShortNameIndex = lnl + 1;
      strcpy(&ofp->FileName[ofp->ShortNameIndex], fdp->cAlternateFileName);
   } else {
      ofp->ShortNameIndex = lnl;
   }
   *fl->tail = ofp;
   fl->tail = &ofp->next;
   fl->count += 1;
}

/* PushWorkList pushes a directory name onto the worklist. It allocates
// enough space to append \*.* to the end so the contents can be
// easily read with FintFirstFile. This also frees the OneFile arg
// since it is always the last use.
*/
void
PushWorkList(OneFile * ofp, BOOL print_me)
{
   DWORD tlen = strlen(ofp->FileName) + 6;
   OneDir * odp = (OneDir *)NTCLImalloc(sizeof(OneDir) + tlen - 1);
   if (! print_me) {
      odp->print_ndx = tlen - 6;
   } else {
      odp->print_ndx = 0;
   }
   strcpy(odp->dirname, ofp->FileName);
   odp->next = worklist.head;
   worklist.head = odp;
   NTCLIfree((void *)ofp);
}

/* PushSubdirWorkList pushes a new directory name on the worklist. The
// name is formed by concatenating the parent directory name with the
// subdirectory name. (If the dir name is just "." or ".\\", then don't
// bother with using it as a prefix). Again, be sure to allocate enough
// extra space for the name so it can have a "\\*.*" appended to it.
*/
void
PushSubdirWorkList(OneDir * odpin, OneFile * subdir)
{
   DWORD prflen = strlen(odpin->dirname);
   DWORD sublen = strlen(subdir->FileName);
   DWORD totlen;
   OneDir * odp;
   DWORD subndx;

   if ((prflen <= 2) &&
       (odpin->dirname[0] == '.') &&
       ((odpin->dirname[1] == '\\') || (odpin->dirname[1] == '\0'))) {
      prflen = 0;
   }
   totlen = prflen + 1 + sublen + 6;
   odp = (OneDir *)NTCLImalloc(sizeof(OneDir) + totlen - 1);
   if (prflen == 0) {
      strcpy(odp->dirname, subdir->FileName);
      subndx = 0;
   } else {
      strcpy(odp->dirname, odpin->dirname);
      if (odp->dirname[prflen-1] != '\\') {
         strcat(odp->dirname, "\\");
      }
      subndx = strlen(odp->dirname);
      strcat(odp->dirname, subdir->FileName);
   }
   if (odpin->dirname[odpin->print_ndx] == '\0') {
      odp->print_ndx = subndx;
   } else {
      odp->print_ndx = odpin->print_ndx;
   }
   odp->next = worklist.head;
   worklist.head = odp;
}

/********************************************************** utility routines */

FileList *
ReadOneDirectory(OneDir * odp)
{
   DWORD len = strlen(odp->dirname);
   HANDLE h;
   WIN32_FIND_DATA fd;
   FileList * d = AllocFileList();

   if (odp->dirname[len - 1] == '\\') { /* setup FindFirstFile call */
      strcat(odp->dirname, "*.*");
   } else {
      strcat(odp->dirname, "\\*.*");
   }
   h = FindFirstFile(odp->dirname, &fd);
   if (h != INVALID_HANDLE_VALUE) {
      for ( ; ; ) {
         /* Here we have a new file name to consider. We need to filter it
         // through any ignore patterns and decide if we are supposed to
         // list it due to it being hidden or special in some way, but if we
         // like this file, it gets added to the list.
         */
         BOOL show_file = TRUE;
         IgnoreList * ilp;

         if ((! show_dot_dot) && (fd.cFileName[0] == '.')) {
            if ((strcmp(fd.cFileName, ".") == 0) ||
                (strcmp(fd.cFileName, "..") == 0)) {
               show_file = FALSE;
            }
         }
         if ((! show_hidden) &&
             (fd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)) {
            show_file = FALSE;
         }
         /* Check file against ignore patterns
         */
         for (ilp = ignore_list; ilp != NULL; ilp = ilp->next) {
            char * name = fd.cFileName;
            if (show_short_names) {
               if (fd.cAlternateFileName[0] != '\0') {
                  name = fd.cAlternateFileName;
               }
            }
            if (NTCLIWildMatch(name, ilp->pattern, FALSE)) {
               show_file = FALSE;
               break;
            }
         }
         if (show_file) {
            PutFindDataOnList(d, &fd);
         }
         if (! FindNextFile(h, &fd)) {
            break;
         }
      }
      FindClose(h);
   }
   odp->dirname[len] = '\0';            /* restore orig name */
   current_directory_name = odp->dirname;
   return d;
}

static void
usage()
{
   fputs("\
ls for Windows NT\n\
usage: ls [options] [files...]\n\
\n\
-a  do not hide hidden or system files or the . and .. dir names\n\
-b  escape unprintable characters to octal\n\
-c  sort on and/or list create time of file\n\
-d  show directory names, not directory contents\n\
-h  print help\n\
-j  convert names to all lower case for printing and sorting\n\
-l  use a long listing format\n\
-m  fill width with a comma separated list of entries\n\
-o  show both names of file in -l listing (for files with MS-DOS aliases)\n\
-p  append a character indicating kind of non-executable files\n\
-q  escape unprintable characters to a question mark\n\
-r  reverse order while sorting\n\
-t  sort on and/or list modification time of file\n\
-u  sort on and/or list access time of file\n\
-w  give argument to specify screen width\n\
-x  list across lines rather than down columns\n\
-A  do not list the . and .. entries, but show other hidden files\n\
-C  list entries by columns\n\
-D  use DOS format for names with \\ as dir separator rather than /\n\
-F  append a character indicating kind of files\n\
-G  turn on display of group information in -l listing\n\
-I  give argument to specify wild card pattern to be ignored\n\
-H  print help\n\
-J  convert names to all upper case for printing and sorting\n\
-M  use the 8.3 format MS-DOS names of files\n\
-N  do not quote entry names\n\
-O  turn on display of owner information in -l listing\n\
-Q  enclose entry names in double quotes\n\
-R  list subdirectories recursively\n\
-S  sort by file size\n\
-U  list entries in unspecified order without any sorting\n\
-X  sort alphabetically by entry extension\n\
-1  list one file per line\n\
\n\
default argument is current directory.\n\
\n\
default format is name only by columns if output is to a console, or\n\
in a single column if output is not to a console.\n", stdout);
}

/*
avail option chars (these don't conflict with the GNU ls and aren't used
by this ls (yet :-)).

-e -v -y -z -E -K -P -V -W -Y -Z
*/

static void
RememberIgnorePattern(char * pat)
{
   IgnoreList * ilp = (IgnoreList *)NTCLImalloc(sizeof(IgnoreList));
   ilp->next = ignore_list;
   ilp->pattern = pat;
   ignore_list = ilp;
}

/****************************************************** Printing and Sorting */

/* NamePrintWidth calculates the printable width of just a name string
*/
DWORD
NamePrintWidth(TCHAR * name)
{
   DWORD rlen = 0;

   if (quote_names) {
      rlen += 2;
   }
   if (escape_unprintable) {
      TCHAR c;
      while ((c = *name++) != '\0') {
         if (isprint(c)) {
            rlen += 1;
         } else {
            rlen += 4; /* backslash and 3 octal digits */
         }
      }
   } else {
      rlen += strlen(name);
   }
   return rlen;
}

BOOL
IsExecutable(OneFile * ofp)
{
   DWORD len = strlen(ofp->FileName);
   if (len >= 5) {
      TCHAR * suffix = &ofp->FileName[len - 4];
      if (*suffix++ == '.') {
         if ((_stricmp(suffix, "exe") == 0) ||
             (_stricmp(suffix, "cmd") == 0) ||
             (_stricmp(suffix, "bat") == 0) ||
             (_stricmp(suffix, "com") == 0)) {
            return TRUE;
         }
      }
   }
   return FALSE;
}

void
NeedPrintBuf(DWORD rlen)
{
   if (rlen > print_buf_len) {
      print_buf_len *= 2;
      if (rlen > print_buf_len) {
         print_buf_len = rlen * 2;
      }
      if (print_buf_len < 500) {
         print_buf_len = 500;
      }
      print_buf = NTCLIrealloc(print_buf, print_buf_len);
   }
}

/* PrintWidth calculates the width required to print a file name
// according to the options that control how individual names are printed.
// it also (if necessary) re-allocates the static buffer used to format
// each individual name so it will have enough space to hold the name.
*/
DWORD
PrintWidth(OneFile * ofp)
{
   DWORD rlen = 0;
   if (long_format) {
      /* The long format width includes the width of the attribute
      // owner, group, date and time info as well as the name and
      // if short names are being printed as well, the name of the
      // short alias (if any).
      */
      rlen += 50; /* guess for now */
      if (show_group) {
         rlen += 10; /* another guess */
      }
      rlen += NamePrintWidth(ofp->FileName);
      if ((show_short_names) && (ofp->FileName[ofp->ShortNameIndex] != '\0')) {
         rlen += 4; /* length of " or " */
         rlen += NamePrintWidth(&ofp->FileName[ofp->ShortNameIndex]);
      }
   } else {
      /* In short format, we only care about the length of the name
      */
      if ((show_short_names) && (ofp->FileName[ofp->ShortNameIndex] != '\0')) {
         rlen += NamePrintWidth(&ofp->FileName[ofp->ShortNameIndex]);
      } else {
         rlen += NamePrintWidth(ofp->FileName);
      }
      if (comma_separate) {
         rlen += 1;
      }
   }
   if ((flag_dirs) && (ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      rlen += 1;
   } else if (flag_executables) {
      /* What makes a file executable is its name, not its attributes.
      // Check for .bat, .com, .exe, or .cmd and call them executable
      */
      if (IsExecutable(ofp)) {
         rlen += 1;
      }
   }
   NeedPrintBuf(rlen);
   return rlen;
}

/* FormatOneName fills in print_buf with the printable info for one file
// name entry (only the short name version is written yet).
*/
void
FormatOneName(OneFile * ofp)
{
   TCHAR * np;
   char * dp = print_buf;
   TCHAR c;

   if ((show_short_names) && (ofp->FileName[ofp->ShortNameIndex] != '\0')) {
      np = &ofp->FileName[ofp->ShortNameIndex];
   } else {
      np = ofp->FileName;
   }
   if (quote_names) {
      *dp++ = '\"';
   }
   while ((c = *np++) != '\0') {
      if (isprint(c)) {
         if (convert_to_lower) {
            c = tolower(c);
         } else if (upcase_names) {
            c = toupper(c);
         }
         if ((! use_backslash_sep) && (c == '\\')) {
            c = '/';
         }
         *dp++ = c;
      } else {
         if (escape_unprintable) {
            char temp[5];
            sprintf(temp,"\\%03o",c);
            *dp++ = temp[0];
            *dp++ = temp[1];
            *dp++ = temp[2];
            *dp++ = temp[3];
         } else if (question_unprintable) {
            *dp++ = '?';
         } else {
            *dp++ = c;
         }
      }
   }
   if ((flag_dirs) && (ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      if (use_backslash_sep) {
         *dp++ = '\\';
      } else {
         *dp++ = '/';
      }
   } else if (flag_executables) {
      /* What makes a file executable is its name, not its attributes.
      // Check for .bat, .com, .exe, or .cmd and call them executable
      */
      if (IsExecutable(ofp)) {
         *dp++ = '*';
      }
   }
   if (quote_names) {
      *dp++ = '\"';
   }
   *dp = '\0';
}

int
NocaseSortByExtension(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   TCHAR * na = ofa->FileName;
   TCHAR * nb = ofb->FileName;
   TCHAR * nax;
   TCHAR * nbx;
   int rval;
   if (show_short_names) {
      if (ofa->FileName[ofa->ShortNameIndex] != '\0') {
         na = &ofa->FileName[ofa->ShortNameIndex];
      }
      if (ofb->FileName[ofb->ShortNameIndex] != '\0') {
         nb = &ofb->FileName[ofb->ShortNameIndex];
      }
   }
   nax = strrchr(na, '.');
   if (nax == NULL) {
      nax = ".";
   }
   nbx = strrchr(nb, '.');
   if (nbx == NULL) {
      nbx = ".";
   }
   rval = _stricmp(nax, nbx);
   if (rval == 0) {
      rval = _stricmp(na, nb);
   }
   return rval;
}

int
SortByExtension(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   TCHAR * na = ofa->FileName;
   TCHAR * nb = ofb->FileName;
   TCHAR * nax;
   TCHAR * nbx;
   int rval;
   if (show_short_names) {
      if (ofa->FileName[ofa->ShortNameIndex] != '\0') {
         na = &ofa->FileName[ofa->ShortNameIndex];
      }
      if (ofb->FileName[ofb->ShortNameIndex] != '\0') {
         nb = &ofb->FileName[ofb->ShortNameIndex];
      }
   }
   nax = strrchr(na, '.');
   if (nax == NULL) {
      nax = ".";
   }
   nbx = strrchr(nb, '.');
   if (nbx == NULL) {
      nbx = ".";
   }
   rval = strcmp(nax, nbx);
   if (rval == 0) {
      rval = strcmp(na, nb);
   }
   return rval;
}

int
SortBySize(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   if (ofa->FileSizeHigh > ofb->FileSizeHigh) {
      return -1;
   } else if (ofa->FileSizeHigh < ofb->FileSizeHigh) {
      return 1;
   } else {
      if (ofa->FileSizeLow > ofb->FileSizeLow) {
         return -1;
      } else if (ofa->FileSizeLow < ofb->FileSizeLow) {
         return 1;
      }
   }
   return 0;
}

int
SortByAccessTime(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   FILETIME * fta = &ofa->LastAccessTime;
   FILETIME * ftb = &ofb->LastAccessTime;
   if ((fta->dwLowDateTime == 0) && (fta->dwHighDateTime == 0)) {
      fta = &ofa->LastWriteTime;
   }
   if ((ftb->dwLowDateTime == 0) && (ftb->dwHighDateTime == 0)) {
      ftb = &ofb->LastWriteTime;
   }
   return CompareFileTime(ftb, fta);
}

int
SortByCreateTime(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   FILETIME * fta = &ofa->CreationTime;
   FILETIME * ftb = &ofb->CreationTime;
   if ((fta->dwLowDateTime == 0) && (fta->dwHighDateTime == 0)) {
      fta = &ofa->LastWriteTime;
   }
   if ((ftb->dwLowDateTime == 0) && (ftb->dwHighDateTime == 0)) {
      ftb = &ofb->LastWriteTime;
   }
   return CompareFileTime(ftb, fta);
}

int
SortByModtime(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   FILETIME * fta = &ofa->LastWriteTime;
   FILETIME * ftb = &ofb->LastWriteTime;
   return CompareFileTime(ftb, fta);
}

int
NocaseSortByName(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   TCHAR * na = ofa->FileName;
   TCHAR * nb = ofb->FileName;
   if (show_short_names) {
      if (ofa->FileName[ofa->ShortNameIndex] != '\0') {
         na = &ofa->FileName[ofa->ShortNameIndex];
      }
      if (ofb->FileName[ofb->ShortNameIndex] != '\0') {
         nb = &ofb->FileName[ofb->ShortNameIndex];
      }
   }
   return _stricmp(na, nb);
}

int
SortByName(const void * a, const void * b)
{
   OneFile * ofa = *((OneFile **)a);
   OneFile * ofb = *((OneFile **)b);
   TCHAR * na = ofa->FileName;
   TCHAR * nb = ofb->FileName;
   if (show_short_names) {
      if (ofa->FileName[ofa->ShortNameIndex] != '\0') {
         na = &ofa->FileName[ofa->ShortNameIndex];
      }
      if (ofb->FileName[ofb->ShortNameIndex] != '\0') {
         nb = &ofb->FileName[ofb->ShortNameIndex];
      }
   }
   return strcmp(na, nb);
}

int
ReverseNocaseSortByExtension(const void * a, const void * b)
{
   return NocaseSortByExtension(b,a);
}

int
ReverseSortByExtension(const void * a, const void * b)
{
   return SortByExtension(b,a);
}

int
ReverseSortBySize(const void * a, const void * b)
{
   return SortBySize(b,a);
}

int
ReverseSortByAccessTime(const void * a, const void * b)
{
   return SortByAccessTime(b,a);
}

int
ReverseSortByCreateTime(const void * a, const void * b)
{
   return SortByCreateTime(b,a);
}

int
ReverseSortByModtime(const void * a, const void * b)
{
   return SortByModtime(b,a);
}

int
ReverseNocaseSortByName(const void * a, const void * b)
{
   return NocaseSortByName(b,a);
}

int
ReverseSortByName(const void * a, const void * b)
{
   return SortByName(b, a);
}

struct sixtyfour {
   DWORD low;
   DWORD high;
};

/* DivideByTen takes a 64 bit unsigned int (stored in a struct sixtyfour)
// and divides it by 10, returning the remainder and modifying the
// struct in place.
*/
DWORD
DivideByTen(struct sixtyfour * sf)
{
   DWORD rem;

   if (sf->high == 0) {
      rem = sf->low % 10;
      sf->low = sf->low / 10;
   } else {
      unsigned long d[4];

      d[0] = (sf->high >> 16) & 0xffff;
      d[1] = sf->high & 0xffff;
      d[2] = (sf->low >> 16) & 0xffff;
      d[3] = sf->low & 0xffff;

      rem = d[0] % 10;
      d[0] = d[0] / 10;
      d[1] |= (rem << 16);
      rem = d[1] % 10;
      d[1] = d[1] / 10;
      d[2] |= (rem << 16);
      rem = d[2] % 10;
      d[2] = d[2] / 10;
      d[3] |= (rem << 16);
      rem = d[3] % 10;
      d[3] = d[3] / 10;

      sf->high = ((d[0] << 16) | d[1]);
      sf->low = ((d[2] << 16) | d[3]);
   }
   return rem;
}

static char * months [] = {
   "Jan", "Feb", "Mar", "Apr", "May", "Jun",
   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

char * sdbuf = NULL;
DWORD sdbuflen = 0;

void
FormatListLong(FileList * fl)
{
   OneFile * ofp;
   char digits[30];
   int diglen;
   struct sixtyfour sf;
   FILETIME * ft;
   FILETIME lft;
   SYSTEMTIME stime;
   char * mn;

   for (ofp = fl->head; ofp != NULL; ofp = ofp->next) {
      /* Print mode flags first on line */

      if (ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
         putc('d', stdout);
      } else {
         putc('-', stdout);
      }
      putc('r', stdout);
      if (ofp->FileAttributes & FILE_ATTRIBUTE_READONLY) {
         putc('-', stdout);
      } else {
         putc('w', stdout);
      }
      if (IsExecutable(ofp) ||
          (ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
         putc('x', stdout);
      } else {
         putc('-', stdout);
      }
      if (ofp->FileAttributes & FILE_ATTRIBUTE_HIDDEN) {
         putc('h', stdout);
      } else {
         putc('-', stdout);
      }
      if (ofp->FileAttributes & FILE_ATTRIBUTE_SYSTEM) {
         putc('s', stdout);
      } else {
         putc('-', stdout);
      }
      if (ofp->FileAttributes & FILE_ATTRIBUTE_ARCHIVE) {
         putc('a', stdout);
      } else {
         putc('-', stdout);
      }
      if (ofp->FileAttributes & FILE_ATTRIBUTE_TEMPORARY) {
         putc('t', stdout);
      } else {
         putc('-', stdout);
      }
      putc(' ', stdout);

      /* Go through horrible contortions to print the damn owner and group
      // names (Sheesh!).
      */
      if ((show_owner) || (show_group)) {
         /* First, need to get the full path name of the file to request
         // the security descriptor for it. Use the print_buf as a handy
         // temp space for this.
         */
         DWORD dirlen = ((current_directory_name == NULL) ?
                            0 :
                            strlen(current_directory_name));
         DWORD namelen = strlen(ofp->FileName);
         DWORD totlen = dirlen + namelen + 2;
         DWORD sdbufneeded;
         DWORD gotsec = TRUE;

         NeedPrintBuf(totlen);
         if (current_directory_name != NULL) {
            strcpy(print_buf, current_directory_name);
            if (print_buf[dirlen-1] != '\\') {
               print_buf[dirlen++] = '\\';
            }
         }
         strcpy(&print_buf[dirlen], ofp->FileName);
         if (sdbuflen < 1000) {
            sdbuflen = 1000;
            sdbuf = (char *)NTCLIrealloc((void *)sdbuf, sdbuflen);
         }
         if (! GetFileSecurity(print_buf,
                               OWNER_SECURITY_INFORMATION |
                                  GROUP_SECURITY_INFORMATION,
                               (PSECURITY_DESCRIPTOR)sdbuf,
                               sdbuflen,
                               &sdbufneeded)) {
            if (sdbufneeded > sdbuflen) {
               sdbuflen = sdbufneeded;
               sdbuf = (char *)NTCLIrealloc((void *)sdbuf, sdbuflen);
               if (! GetFileSecurity(print_buf,
                                     OWNER_SECURITY_INFORMATION |
                                        GROUP_SECURITY_INFORMATION,
                                     (PSECURITY_DESCRIPTOR)sdbuf,
                                     sdbuflen,
                                     &sdbufneeded)) {
                  gotsec = FALSE;
               }
            }
         }
         if (gotsec) {
            if (show_owner) {
               BOOL who_cares;
               PSID ownsid;
               TCHAR accname[100];
               DWORD accsize = sizeof(accname);
               TCHAR domname[100];
               DWORD domsize = sizeof(domname);
               SID_NAME_USE snu;
               char * owname;

               if (GetSecurityDescriptorOwner((PSECURITY_DESCRIPTOR)sdbuf,
                                              &ownsid,
                                              &who_cares) &&
                   LookupAccountSid(NULL,
                                    ownsid,
                                    accname,
                                    &accsize,
                                    domname,
                                    &domsize,
                                    &snu)) {
                  owname = accname;
               } else {
                  owname = " no owner ";
               }
               printf("%-14s ", owname);
            }
            if (show_group) {
               BOOL who_cares;
               PSID grpsid;
               TCHAR accname[100];
               DWORD accsize = sizeof(accname);
               TCHAR domname[100];
               DWORD domsize = sizeof(domname);
               SID_NAME_USE snu;
               char * grpname;

               if (GetSecurityDescriptorGroup((PSECURITY_DESCRIPTOR)sdbuf,
                                              &grpsid,
                                              &who_cares) &&
                   LookupAccountSid(NULL,
                                    grpsid,
                                    accname,
                                    &accsize,
                                    domname,
                                    &domsize,
                                    &snu)) {
                  grpname = accname;
               } else {
                  grpname = " no group ";
               }
               printf("%-14s ", grpname);
            }
         } else {
            printf("no security info ");
         }
      }

      /* Print size (which, unfortunately for us, can be a 64 bit number) */
      sf.high = ofp->FileSizeHigh;
      sf.low = ofp->FileSizeLow;
      diglen = 0;
      do {
         digits[diglen++] = (char)('0' + DivideByTen(&sf));
      } while (! ((sf.high == 0) && (sf.low == 0)));
      if (diglen < 8) {
         int count = 8 - diglen;
         while (count-- > 0) {
            putc(' ', stdout);
         }
      }
      while (--diglen >= 0) {
         putc(digits[diglen], stdout);
      }
      putc(' ', stdout);

      /* Print time */
      if (use_access_time) {
         ft = &ofp->LastAccessTime;
      } else if (use_create_time) {
         ft = &ofp->CreationTime;
      } else {
         ft = &ofp->LastWriteTime;
      }
      if ((ft->dwLowDateTime == 0) && (ft->dwHighDateTime == 0)) {
         ft = &ofp->LastWriteTime;
      }
      if (FileTimeToLocalFileTime(ft, &lft) &&
          FileTimeToSystemTime(&lft, &stime)) {
         /* yy mmm dd hh:mm:ss */
         if ((stime.wMonth >= 1) && (stime.wMonth <= 12)) {
            mn = months[stime.wMonth - 1];
         } else {
            mn = "Huh";
         }
         printf("%2d %s %2d %02d:%02d:%02d",
            stime.wYear % 100,
            mn,
            stime.wDay,
            stime.wHour,
            stime.wMinute,
            stime.wSecond);
      } else {
         /* Some sort of time conversion error... */
         fputs("time convert error", stdout);
      }
      putc(' ', stdout);

      /* Print primary name first
      */
      PrintWidth(ofp);
      FormatOneName(ofp);
      fputs(print_buf, stdout);
      /* Then secondary name if requested (and there is one) (temporarily
      // invert the show_short_names option to coerce the utility routine to
      // format the desired name).
      */
      if (show_both_names &&
          (ofp->FileName[ofp->ShortNameIndex] != '\0')) {
         show_short_names = ! show_short_names;
         PrintWidth(ofp);
         FormatOneName(ofp);
         fputs(" or ", stdout);
         fputs(print_buf, stdout);
         show_short_names = ! show_short_names;
      }

      /* We are done */
      putc('\n', stdout);
   }
}

/* FormatListComma lists files packed across the line separated by commas.
*/
void
FormatListComma(FileList * fl)
{
   OneFile * ofp;
   DWORD cursize = 0;
   for (ofp = fl->head; ofp != NULL; ofp = ofp->next) {
      DWORD thisw = PrintWidth(ofp);
      if ((cursize > 0) && ((cursize + thisw + 2) >= screen_width)) {
         putc('\n', stdout);
         cursize = 0;
      }
      FormatOneName(ofp);
      fputs(print_buf, stdout);
      cursize += strlen(print_buf);
      if (ofp->next != NULL) {
         putc(',', stdout);
         cursize += 1;
      } else {
         putc('\n', stdout);
      }
   }
}

/* FormatListAcross lists the files in order across the screen rather than
// down the columns.
*/
void
FormatListAcross(FileList * fl)
{
   OneFile * ofp;
   DWORD maxw = 0;
   DWORD numcols;
   DWORD i;

   for (ofp = fl->head; ofp != NULL; ofp = ofp->next) {
      DWORD thisw = PrintWidth(ofp);
      if (thisw > maxw) {
         maxw = thisw;
      }
   }
   maxw += 1;
   maxw = ((maxw + 7) / 8) * 8;
   numcols = screen_width / maxw;
   if (numcols <= 0) {
      numcols = 1;
   }
   i = 0;
   for (ofp = fl->head; ofp != NULL; ofp = ofp->next) {
      ++i;
      FormatOneName(ofp);
      fputs(print_buf, stdout);
      if (i >= numcols) {
         putc('\n', stdout);
         i = 0;
      } else if (ofp->next != NULL) {
         DWORD k = maxw - strlen(print_buf);
         while (k-- > 0) {
            putc(' ', stdout);
         }
      } else {
         putc('\n', stdout);
      }
   }
}

/* FormatListOneColumn is the simplest listing routine, generating a single
// column listing.
*/
void
FormatListOneColumn(FileList * fl)
{
   OneFile * ofp;

   for (ofp = fl->head; ofp != NULL; ofp = ofp->next) {
      PrintWidth(ofp);
      FormatOneName(ofp);
      fputs(print_buf, stdout); putc('\n', stdout);
   }
}

/* FormatListDown formats a list by columns with names listed in order down
// columns (not across).
*/
void
FormatListDown(FileList * fl)
{
   OneFile * ofp;
   OneFile ** ofl;
   DWORD i = 0;
   DWORD maxw = 0;
   DWORD numcols;
   DWORD numrows;
   DWORD j,k;

   ofp = fl->head;
   ofl = (OneFile **)NTCLImalloc(sizeof(OneFile *) * fl->count);
   while (ofp != NULL) {
      DWORD thisw;

      ofl[i++] = ofp;
      thisw = PrintWidth(ofp);
      if (thisw > maxw) {
         maxw = thisw;
      }
      ofp = ofp->next;
   }
   maxw += 1;
   maxw = ((maxw + 7) / 8) * 8;
   numcols = screen_width / maxw;
   if (numcols <= 0) {
      numcols = 1;
   }
   numrows = fl->count / numcols;
   if ((fl->count % numcols) != 0) {
      numrows += 1;
   }
   for (i = 0; i < numrows; ++i) {
      DWORD prefix = 0;
      for (j = 0; j < numcols; ++j) {
         DWORD ndx = numrows * j + i;
         if (ndx < fl->count) {
            for (k = 0; k < prefix; ++k) {
               putc(' ', stdout);
            }
            FormatOneName(ofl[ndx]);
            prefix = maxw - strlen(print_buf);
            fputs(print_buf, stdout);
         }
      }
      putc('\n', stdout);
   }
}

/* FormatList makes all the decisions about how to sort and print based on
// the runtime options (actually, it would probably make sense to do this
// once and save pointers to the correct routines, but it only makes a
// difference when doing recursive lists, so it isn't that big a deal).
*/
void
FormatList(FileList * fl)
{
   /* First sort the list */

   if (! do_not_sort) {
      if (sort_by_extension) {
         if (convert_to_lower || upcase_names) {
            if (reverse_sort) {
               SortFileList(fl, ReverseNocaseSortByExtension);
            } else {
               SortFileList(fl, NocaseSortByExtension);
            }
         } else {
            if (reverse_sort) {
               SortFileList(fl, ReverseSortByExtension);
            } else {
               SortFileList(fl, SortByExtension);
            }
         }
      } else if (sort_by_size) {
         if (reverse_sort) {
            SortFileList(fl, ReverseSortBySize);
         } else {
            SortFileList(fl, SortBySize);
         }
      } else if (use_access_time) {
         if (reverse_sort) {
            SortFileList(fl, ReverseSortByAccessTime);
         } else {
            SortFileList(fl, SortByAccessTime);
         }
      } else if (use_create_time) {
         if (reverse_sort) {
            SortFileList(fl, ReverseSortByCreateTime);
         } else {
            SortFileList(fl, SortByCreateTime);
         }
      } else if (use_modtime) {
         if (reverse_sort) {
            SortFileList(fl, ReverseSortByModtime);
         } else {
            SortFileList(fl, SortByModtime);
         }
      } else {
         if (convert_to_lower || upcase_names) {
            if (reverse_sort) {
               SortFileList(fl, ReverseNocaseSortByName);
            } else {
               SortFileList(fl, NocaseSortByName);
            }
         } else {
            if (reverse_sort) {
               SortFileList(fl, ReverseSortByName);
            } else {
               SortFileList(fl, SortByName);
            }
         }
      }
   }

   /* Then print it */

   if (long_format) {
      FormatListLong(fl);
   } else if (comma_separate) {
      FormatListComma(fl);
   } else if (list_by_columns) {
      if (list_across) {
         FormatListAcross(fl);
      } else {
         FormatListDown(fl);
      }
   } else {
      FormatListOneColumn(fl);
   }
}

/********************************************************** The main program */

int
main(int argc, char ** argv)
{
   FileList * args;
   OneDir * od;
   OneDir * odp;
   OneFile * ofp;
   HANDLE console;
   CONSOLE_SCREEN_BUFFER_INFO bufinfo;
   OneFile * tempof = NULL;
   DWORD tempoflen = 0;
   NTCLIOptinfo oi;
   char * onearg;

   /* I'll be trying to wildcard expand things which might be on removable
   // devices, so tell the system to take a hike when it comes to displaying
   // error pop-up window's and just return an error code from the system
   // service instead.
   */
   SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOOPENFILEERRORBOX);

   /* initialize default listing mode depending on kind of output device. If
   // it looks like a console, then use by columns as the default and get
   // width from console width, otherwise don't list by columns and mark
   // width as unknown.
   */
   console = GetStdHandle(STD_OUTPUT_HANDLE);
   if ((console != INVALID_HANDLE_VALUE) &&
       GetConsoleScreenBufferInfo(console, &bufinfo)) {
      screen_width = (DWORD) bufinfo.dwSize.X;
      list_by_columns = TRUE;
   } else {
      list_by_columns = FALSE;
   }
   worklist.head = NULL;
   oi = NTCLIInitOptinfo(argc, argv, "w:I:abcdhHjlmopqrtuxACDFGJMNOQRSUX1");
   progname = NTCLIGetProgname(oi);
   for ( ; ; ) {
      int opt = NTCLIGetNextOption(oi);
      if (opt == -1) {
         char * msg = NTCLIErrorString(oi);
         if (msg != NULL) {
            fprintf(stderr, "%s\n", msg);
            exit(2);
         } else {
            break;
         }
      } else {
         switch(opt) {
         case 'a':
            show_hidden = TRUE;
            show_dot_dot = TRUE;
            break;
         case 'b':
            escape_unprintable = TRUE;
            break;
         case 'c':
            use_create_time = TRUE;
            break;
         case 'd':
            show_dir_contents = FALSE;
            break;
         case 'h':
         case 'H':
            usage();
            exit(0);
         case 'j':
            convert_to_lower = TRUE;
            break;
         case 'l':
            long_format = TRUE;
            break;
         case 'm':
            list_across = TRUE;
            comma_separate = TRUE;
            break;
         case 'o':
            show_both_names = TRUE;
            break;
         case 'p':
            flag_dirs = TRUE;
            break;
         case 'q':
            question_unprintable = TRUE;
            break;
         case 'r':
            reverse_sort = TRUE;
            break;
         case 't':
            use_modtime = TRUE;
            break;
         case 'u':
            use_access_time = TRUE;
            break;
         case 'w':
            screen_width_string = NTCLIGetLastOptionArg(oi);
            break;
         case 'x':
            list_across = TRUE;
            break;
         case 'A':
            show_hidden = TRUE;
            show_dot_dot = FALSE;
            break;
         case 'C':
            list_by_columns = TRUE;
            break;
         case 'D':
            use_backslash_sep = TRUE;
            break;
         case 'F':
            flag_dirs = TRUE;
            flag_executables = TRUE;
            break;
         case 'G':
            show_group = TRUE;
            break;
         case 'I':
            RememberIgnorePattern(NTCLIGetLastOptionArg(oi));
            break;
         case 'J':
            upcase_names = TRUE;
            break;
         case 'M':
            show_short_names = TRUE;
            break;
         case 'N':
            quote_names = FALSE;
            break;
         case 'O':
            show_owner = TRUE;
            break;
         case 'Q':
            quote_names = TRUE;
            break;
         case 'R':
            recursive_list = TRUE;
            break;
         case 'S':
            sort_by_size = TRUE;
            break;
         case 'U':
            do_not_sort = TRUE;
            break;
         case 'X':
            sort_by_extension = TRUE;
            break;
         case '1':
            list_by_columns = FALSE;
            break;
         }
      }
   }
   if (screen_width_string != NULL) {
      screen_width = (DWORD)atol(screen_width_string);
   }
   /* By waiting until after the options are processed to expand the
   // remainder of the argv list, I can avoid accidentally expanding the -I
   // arguments but still take advantage of wildcards in the file
   // arguments.
   */
   NTCLIExpandWildcards(oi);

   /* Make up initial argument list, using "." as implicit argument if
   // no other argument given.
   */
   args = AllocFileList();
   onearg = NTCLIGetNextPositionalArg(oi);
   if (onearg == NULL) {
      PutStringOnList(args, ".");
   } else {
      PutStringOnList(args, onearg);
      while ((onearg = NTCLIGetNextPositionalArg(oi)) != NULL) {
         PutStringOnList(args, onearg);
      }
   }

   /* Now decide exactly what to do. If I am supposed to list the contents
   // of directories, and there is more than one argument, then go through
   // the argument list, separating out the directories and pushing them on
   // a work list to be processed individually. If there is only one
   // directory on the list, then simply make the arg list into the
   // directory work list. If I am not supposed to list the contents of
   // directories, or there are non-directory arguments, then just format
   // the contents of the arg list.
   */
   if (show_dir_contents) {
      if (args->count > 1) {
         /* With more than one arg, there will be ambiguities if I don't
         // show the directory name of each directory printed. Go through
         // the list, pushing any directories on the worklist and leaveing
         // any non-directory items on the args list. The print the arg list
         // (if non-empty) and process the directory work list.
         */
         ofp = args->head;
         args->head = NULL;
         args->tail = &args->head;
         args->count = 0;
         while (ofp != NULL) {
            if (ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
               OneFile * nofp = ofp->next;
               ofp->next = NULL;
               PushWorkList(ofp, TRUE);
               ofp = nofp;
            } else {
               OneFile * nofp = ofp->next;
               ofp->next = NULL;
               *(args->tail) = ofp;
               args->tail = &ofp->next;
               args->count += 1;
			   ofp = nofp;
            }
         }
         if (args->count > 0) {
            FormatList(args);
            FreeFileList(args);
         }
         /* Reverse entries in work list to keep them
         // in the same order they were in the arg list
         */
         odp = worklist.head;
         worklist.head = NULL;
         while (odp != NULL) {
            OneDir * ndp = odp->next;
            odp->next = worklist.head;
            worklist.head = odp;
            odp = ndp;
         }
      } else if (args->count == 1) {
         if (args->head->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            /* The single arg was a directory, so the contents need to be
            // printed, but we don't want to show the name of this directory
            // in the resulting list, since there is no ambiguity about a
            // single directory. Just push a directory work list entry, but
            // do not mark the directory name to be printed.
            */
            PushWorkList(args->head, FALSE);
         } else {
            /* The single arg was not a directory, so just print it out.
             */
            FormatList(args);
            FreeFileList(args);
         }
      }
      /* Now loop, processing entries on the directory work list until the
      // work list is empty (if we are recursively listing directories, new
      // entries get pushed on the end of the list as old ones are pulled
      // off the front).
      */
      while ((od = PopWorkList()) != NULL) {
         if (od->dirname[od->print_ndx] != '\0') {
            /* If we are supposed to print part of this directory name, then
            // copy it to a temporary OneFile object so we can pass it to
            // the FormatOneName routine and print it with all the same
            // transformations other names are printed.
            */
            DWORD dlen = strlen(&od->dirname[od->print_ndx]);
            if (dlen > tempoflen) {
               tempoflen *= 2;
               if (dlen > tempoflen) {
                  tempoflen = 2 * dlen;
               }
               if (tempoflen < 500) {
                  tempoflen = 500;
               }
               if (tempof == NULL) {
                  tempof = (OneFile *)NTCLIcalloc(sizeof(OneFile) +
                                                  tempoflen, 1);
                  tempof->FileAttributes |= FILE_ATTRIBUTE_DIRECTORY;
               } else {
                  tempof = (OneFile *)NTCLIrealloc(tempof,
                                                   sizeof(OneFile) + tempoflen);
               }
            }
            strcpy(tempof->FileName, &od->dirname[od->print_ndx]);
            tempof->ShortNameIndex = dlen;
            PrintWidth(tempof);
            FormatOneName(tempof);
            putc('\n', stdout); fputs(print_buf, stdout); putc('\n', stdout);
         }
         args = ReadOneDirectory(od);
         FormatList(args);
         if (recursive_list) {
            /* If we are supposed to list sub-directories as well, then
            // push any subdirectories on the list here (except . and ..).
            */
            ReverseFileList(args);
            ofp = args->head;
            while (ofp != NULL) {
               if ((ofp->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
                   (strcmp(ofp->FileName, ".") != 0) &&
                   (strcmp(ofp->FileName, "..") != 0)) {
                  PushSubdirWorkList(od, ofp);
               }
               ofp = ofp->next;
            }
         }
         FreeFileList(args);
         NTCLIfree((void *)od);
      }
   } else {
      FormatList(args);
      FreeFileList(args);
   }
   return 0;
}
