
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

/* The ntcli.h header defines useful functions that many of the NT
 * Command Line Interface utilities share.
 */

#ifndef ntcli_h
#define ntcli_h

#include <stdlib.h>

#ifndef DllExport
#define DllExport    __declspec( dllexport )
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* The following version is used part of the registry key. It will only
// change when the format of the registry info changes.
*/
#define NTCLI_VERSION   "1.0"

/* "safe" versions of memory allocation routines. Exit with a message if
 * out of memory, and allow freeing of NULL pointers.
*/
DllExport void * NTCLImalloc(size_t);
DllExport void * NTCLIcalloc(size_t, size_t);
DllExport void * NTCLIrealloc(void *, size_t);
DllExport void   NTCLIfree(void *);

/* Argument vector processing routines.
*/

struct NTCLIOptinfo_struct;
typedef struct NTCLIOptinfo_struct * NTCLIOptinfo;

/* Call NTCLIInitOptinfo first to initialize processing. Pass it the argc
// and argv information from the command arguments, as well as an option
// description string like getopt() normally takes.
//
// Return value is an anonymous pointer which should be passed as the first
// argument to any of the remaining routines.
//
// The program base name is extracted from the argv[0] entry and used to
// lookup any user specific default options for this program in the
// registry. These defaults (if they are defined) are returned first by the
// NTCLIGetNextOption routine.
//
// The registry is also consulted to determine if wildcard expansion should
// accept / as a directory separator as well as \ and if the patterns should
// be case sensitive or not and if hidden files should be seen by
// wildcards.
*/
DllExport NTCLIOptinfo
NTCLIInitOptinfo(int argc, char ** argv, char * optdesc);

/* NTCLIGetProgname - return the base name of the program as extracted from
// argv[0]. This is the name used in the registry lookup.
*/
DllExport char *
NTCLIGetProgname(NTCLIOptinfo oi);

/* NTCLIGetNextOption is called to get the next option letter from the
// command line (very much like getopt). It returns -1 if there are no more
// options or if an error occurs (like an undefined option or missing
// argument).
*/
DllExport int
NTCLIGetNextOption(NTCLIOptinfo oi);

/* NTCLIGetLastOptionArg returns a pointer to the argument string of the
// last option, or NULL if there is no argument.
*/
DllExport char *
NTCLIGetLastOptionArg(NTCLIOptinfo oi);

/* When NTCLIGetNextOption returns -1, this routine can be called to
// determine if there was an error, or if you simply reached the end of the
// arguments (if there was a recoverable error, like unknown argument, it
// also flushes the error information and allows you to continue
// processing).
*/
DllExport char *
NTCLIErrorString(NTCLIOptinfo oi);

/* After you run out of options, you can call NTCLIGetNextPositionalArg to
// return each of the remaining arguments in turn. It returns NULL when
// there are no more arguments.
*/
DllExport char *
NTCLIGetNextPositionalArg(NTCLIOptinfo oi);

/* Like NTCLIGetNextPositionalArg, but only peeks ahead to see if there is
 * one, does not actually consume an argument.
 */
DllExport char *
NTCLIPeekNextPositionalArg(NTCLIOptinfo oi);

/* NTCLIExpandWildcards may be called at any time after initializing
// argument processing. It does wildcard expansion on any remaining
// arguments that have not yet been processed. If a wildcard cannot be
// expanded, then it remains in the arg list in its original form.
//
// The wildcard expansion supported by this routine is much more advanced
// that the "normal" expansion. It supports character ranges using [] (where
// ! may be used as the first character to invert the range), and it allows
// multiple wildcards in different components of the name (including the
// leading filesystem letter component), so that wildcards such as
//
// [cde]:\*\lib\*.[ch]
//
// are legal and will be expanded "as expected". (This one, for instance
// would expand to the complete list of all .c and .h files in any
// directories 2nd from the top named 'lib' on any of the c:, d:, or e:
// filesystems).
*/
DllExport void
NTCLIExpandWildcards(NTCLIOptinfo oi);

/* When you are sure you are done with argument processing, you can call this
// to free up the space.
*/
DllExport void
NTCLIFreeOptinfo(NTCLIOptinfo oi);

/* Call NTCLIWildMatch to check a string against a pattern. Returns
 * TRUE if the string matches the pattern, falso otherwise. If the
 * case_sensitive is TRUE, then the matching is case sensistive,
 * otherwise case is ignored (a matches A, etc).
 */
DllExport int
NTCLIWildMatch(char *string, char *pattern, int case_sensitive);

#ifdef __cplusplus
}
#endif

#endif /* ntcli_h */
