
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

/* Implementation for fancy command option processor for Windows NT.  (Well,
// it's not too fancy - it doesn't allow long option names like the GNU
// getopt, but boy does it ever allow fancy wildcard expansion!).
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include "ntcli.h"

/* The arg vectors are turned into lists of argv_entry structs */

struct argv_entry {
   struct argv_entry * next;    /* pointer to next entry */
   int                 scanpos; /* next position to scan for wildcards */
   char                name[1]; /* the string holding the name */
};

/* An argv_expansion is a list of argv_entry entries which happens to keep
 * track of the tail of the list (so lists may be quickly appended).
 */
struct argv_expansion {
   struct argv_entry * head;
   struct argv_entry ** tail;
};

/* forward declaration of static functions */

static struct argv_expansion *
ExpandOneArg(
   struct argv_entry * toparg,
   int use_slash,
   int see_hidden,
   int case_sensitive);

static int
OpenVaKey(HKEY root, PHKEY newkeyp, ...);

static int
ReadKeyval(
   HKEY khand,
   char * valname,
   DWORD * keylen,
   void ** keyval,
   DWORD * keytype);

static int
GetFlag(NTCLIOptinfo oi, char * flagname, int default_value);

/* private struct definition */

struct NTCLIOptinfo_struct {
   char *               optdesc;
   int                  use_slash;
   int                  see_hidden;
   int                  case_sensitive;
   struct argv_entry *  argv;
   struct argv_entry ** argv_end;
   char *               registry_string;
   DWORD                registry_len;
   char *               progname;
   int                  state;
   char *               curarg;
   char *               optarg;
   int                  errflg;
   char                 errmsg[80];
};

/* NTCLIInitOptinfo initializes reading the option information for the
 * program. It creates the NTCLIOptinfo struct which tracks the progress of
 * argument parsing and contains the default argument information dug up out
 * of the registry as well as the arguments explicitly passed.
 */
DllExport NTCLIOptinfo
NTCLIInitOptinfo(int argc, char ** argv, char * optdesc)
{
   HANDLE keyhandle;

   NTCLIOptinfo oi =
      (NTCLIOptinfo)NTCLImalloc(sizeof(struct NTCLIOptinfo_struct));
   oi->optdesc = optdesc;
   oi->use_slash = TRUE;
   oi->see_hidden = FALSE;
   oi->case_sensitive = FALSE;
   oi->argv = NULL;
   oi->argv_end = &oi->argv;
   oi->registry_string = NULL;
   oi->registry_len = 0;
   oi->progname = NULL;
   oi->state = 0;
   oi->curarg = NULL;
   oi->errflg = 0;

#ifdef DEBUGGING
   fprintf(stderr, "NTCLIInitOptinfo called, argc = %d, optdesc=%s\n",
           argc, optdesc);
#endif

   /* First dig up the program name out of argv[0] and advance the arg
   // pointer past it.
   */
   if (argc > 0) {
      char * argv0 = argv[0];
      char * endp = argv0 + strlen(argv0);
      char * begp = argv0;
      char * cp = endp;

      while (cp > argv0) {
         char c = *--cp;
         if (c == '.') {
            if (*endp != '.') {
               endp = cp;
            }
         } else if ((c == '\\') || (c == '/')) {
            begp = cp + 1;
            break;
         }
      }
      oi->progname = (char *)NTCLImalloc((endp - begp) + 1);
      cp = oi->progname;
      while (begp < endp) {
         char c = *begp++;
         c = tolower(c);
         *cp++ = c;
      }
      *cp = '\0';
      --argc;
      ++argv;
   }

#ifdef DEBUGGING
   fprintf(stderr, "NTCLIInitOptinfo find program name %s\n", oi->progname);
#endif


   /* Check the registry for the flags controlling argument recognition and
   // expansion (this looks for flags associated with the current program
   // name first, then looks in the global defaults).
   */
   oi->use_slash = GetFlag(oi, "AllowForwardSlash", oi->use_slash);
   oi->see_hidden = GetFlag(oi, "SeeHiddenFiles", oi->see_hidden);
   oi->case_sensitive =
      GetFlag(oi, "CaseSensitiveWildcards", oi->case_sensitive);

#ifdef DEBUGGING
   fprintf(stderr, "NTCLIInitOptinfo: AllowForwardSlash is %s\n",
      oi->use_slash ? "TRUE" : "FALSE");
   fprintf(stderr, "NTCLIInitOptinfo: SeeHiddenFiles is %s\n",
      oi->see_hidden ? "TRUE" : "FALSE");
   fprintf(stderr, "NTCLIInitOptinfo: CaseSensitiveWildcards is %s\n",
      oi->case_sensitive ? "TRUE" : "FALSE");
#endif

   /* Now see if there is a registry entry for:
   //
   // HKEY_CURRENT_USER\Software\NTCLI\<progname>\<version>
   //
   // If there is one, it should have a value named Options which is a
   // multi-string containing the default option list for this program. Pick
   // it up and build an argv vector for it which will be processed prior to
   // the arguments from the command line.
   */
   if (OpenVaKey(HKEY_CURRENT_USER, &keyhandle,
          "Software", "NTCLI", oi->progname, NTCLI_VERSION, NULL)) {
      DWORD keytype;
      if (ReadKeyval(keyhandle, "Options", &oi->registry_len,
                     (void **)&oi->registry_string, &keytype)) {
         if (keytype != REG_MULTI_SZ) {
            NTCLIfree(oi->registry_string);
            oi->registry_string = NULL;
         }
      }
   }
   if (oi->registry_string != NULL) {
      char * rp = oi->registry_string;
      while (*rp != '\0') {
         DWORD rplen = strlen(rp);
         struct argv_entry * av = (struct argv_entry *)
            NTCLImalloc(sizeof(struct argv_entry) + rplen);
         av->next = NULL;
         av->scanpos = 0;
#ifdef DEBUGGING
         fprintf(stderr, "NTCLIInitOptinfo found default option %s\n", rp);
#endif
         strcpy(av->name, rp);
         *(oi->argv_end) = av;
         oi->argv_end = &av->next;
         rp += rplen + 1;
      }
   }

   /* Now copy rest of argv vector to list */

   while (argc > 0) {
      char * ap = argv[0];
      DWORD aplen = strlen(ap);
      struct argv_entry * av = (struct argv_entry *)
         NTCLImalloc(sizeof(struct argv_entry) + aplen);
      av->next = NULL;
      av->scanpos = 0;
#ifdef DEBUGGING
      fprintf(stderr, "NTCLIInitOptinfo passed option %s\n", ap);
#endif
      strcpy(av->name, ap);
      *(oi->argv_end) = av;
      oi->argv_end = &av->next;
      --argc;
      ++argv;
   }

   return oi;
}

/* Retrieve the pointer to the program name parsed out of the argv[0]
 * entry (with path components and .exe suffix stripped off).
 */
DllExport char *
NTCLIGetProgname(NTCLIOptinfo oi)
{
   return oi->progname;
}

/* Get the next option letter, or -1 if all options from command line
 * have been used up or there is an error.
 */
DllExport int
NTCLIGetNextOption(NTCLIOptinfo oi)
{
   if ((oi->state == 1) && (*oi->curarg == '\0')) {
      oi->state = 0;
   }
   switch(oi->state) {
   case 0:
      if (oi->argv != NULL) {
         char * nuarg = oi->argv->name;
         if ((*nuarg == '-') && (*(nuarg + 1) != '\0')) {
            oi->argv = oi->argv->next;
            if (oi->argv == NULL) {
               oi->argv_end = &oi->argv;
            }
            if (strcmp(nuarg, "--") == 0) {
               oi->state = 2;
               return -1;
            } else {
               oi->curarg = nuarg + 1;
               oi->state = 1;
               /* Fall through to case 1 */
            }
         } else {
            oi->state = 2;
            return -1;
         }
      } else {
         oi->state = 2;
         return -1;
      }
   case 1: {
      char optltr = *oi->curarg++;
      char * optdesc = strchr(oi->optdesc, optltr);
      if (optdesc == NULL) {
         /* Flag error by inverting state, record error message */
         oi->state = - oi->state;
         sprintf(oi->errmsg, "%s: Unknown option: -%c",
            oi->progname, optltr);
         oi->errflg = 1;
#ifdef DEBUGGING
         fprintf(stderr, "NTCLIGetNextOption: %s\n", oi->errmsg);
#endif
         return -1;
      } else {
         /* This is an expected option letter, check and see if it
         // has an argument, if so, record it.
         */
         if (optdesc[1] == ':') {
            oi->optarg = oi->curarg;
            oi->state = 0;
            if (*oi->optarg == '\0') {
               if (oi->argv != NULL) {
                  oi->optarg = oi->argv->name;
                  oi->argv = oi->argv->next;
                  if (oi->argv == NULL) {
                     oi->argv_end = &oi->argv;
                  }
               } else {
                  oi->state = 3;
                  sprintf(oi->errmsg, "%s: Missing argument for -%c option",
                     oi->progname, optltr);
                  oi->errflg = 1;
#ifdef DEBUGGING
                  fprintf(stderr, "NTCLIGetNextOption: %s\n", oi->errmsg);
#endif
                  return -1;
               }
            }
         }
         return optltr;
      }
      } break;
   case -1:
      return -1;
   case 2:
   case 3:
      strcpy(oi->errmsg, "No more options");
      oi->errflg = 1;
#ifdef DEBUGGING
      fprintf(stderr, "NTCLIGetNextOption: %s\n", oi->errmsg);
#endif
      return -1;
   }
   return -1;
}

/* Get the argument string associated with the last argument character.
 * Returns NULL if there is no argument or there is an error.
 */
DllExport char *
NTCLIGetLastOptionArg(NTCLIOptinfo oi)
{
   return oi->optarg;
}

/* Get description of last error (or NULL if there was no error).  Clears
 * the error status as a side effect so processing may continue if the error
 * is not fatal.
 */
DllExport char *
NTCLIErrorString(NTCLIOptinfo oi)
{
   if (oi->state < 0) {
      oi->state = - oi->state;
   }
   if (! oi->errflg) {
      return NULL;
   }
   oi->errflg = 0;
   return oi->errmsg;
}

/* After NTCLIGetNextOption returns -1 to indicate an end to option
 * arguments, this routine may be called to return each remaining positional
 * argument until the args are used up, at which time it returns NULL.
 */
DllExport char *
NTCLIGetNextPositionalArg(NTCLIOptinfo oi)
{
   if (oi->argv == NULL) {
      return NULL;
   } else {
      char * rval = oi->argv->name;
      oi->argv = oi->argv->next;
      if (oi->argv == NULL) {
         oi->argv_end = &oi->argv;
      }
      return rval;
   }
}

/* Like NTCLIGetNextPositionalArg, but only peeks ahead to see if there is
 * one, does not actually consume an argument.
 */
DllExport char *
NTCLIPeekNextPositionalArg(NTCLIOptinfo oi)
{
   if (oi->argv == NULL) {
      return NULL;
   } else {
      return oi->argv->name;
   }
}

/* This function may be called at any time after NTCLIInitOptinfo. It
 * immediately expands all remaining arguments which contain wildcard
 * patterns into the multiple files that the wildcard matches (or it leaves
 * the argument alone if the wildcard cannot match any files).
 */
DllExport void
NTCLIExpandWildcards(NTCLIOptinfo oi)
{
   struct argv_entry ** cf;
   struct argv_entry * av;

#ifdef DEBUGGING
   fprintf(stderr, "NTCLIExpandWildcards called...\n");
#endif
   cf = &oi->argv;
   while ((av = *cf) != NULL) {
      struct argv_expansion * ave;

      /* remove arg from list */
      *cf = av->next;
      av->next = NULL;
      if (oi->argv == NULL) {
         oi->argv_end = &oi->argv;
      }

      /* expand it if possible
      */
#ifdef DEBUGGING
      fprintf(stderr, "NTCLIExpandWildcards checking %s\n", av->name);
#endif
      ave = ExpandOneArg(av, oi->use_slash, oi->see_hidden, oi->case_sensitive);

      if (ave == NULL) {
         /* The arg didn't expand, just put it back on list and advance to
         // next arg.
         */
#ifdef DEBUGGING
         fprintf(stderr, "NTCLIExpandWildcards no arg expansion\n");
#endif
         av->next = *cf;
         *cf = av;
         cf = &av->next;
      } else {
         /* The arg did expand, put the entire expansion on the list
         // and free the original arg.
         */
#ifdef DEBUGGING
         fprintf(stderr, "NTCLIExpandWildcards expanded arg replaced.\n");
#endif
         *ave->tail = *cf;
         *cf = ave->head;
         cf = ave->tail;
         NTCLIfree(ave);
      }
   }
}

/* This routine should be called after all argument processing to free up
 * storage.
 */
DllExport void
NTCLIFreeOptinfo(NTCLIOptinfo oi)
{
   NTCLIfree((void *)oi);
}

/* Low level expansion routine */

#define START_RANGE_CHAR '['
#define END_RANGE_CHAR ']'

/* Expand one argument entry. If the arg contains no wildcards, or does
 * contain wildcards, but they do not expand to anything, then return NULL.
 * If there is a valid expansion, then return the argv_expansion list of all
 * the expansions
 */
static struct argv_expansion *
ExpandOneArg(
   struct argv_entry * toparg,
   int use_slash,
   int see_hidden,
   int case_sensitive)
{
   struct argv_entry * todo_list;
   struct argv_entry * av;
   struct argv_entry * done_list = NULL;
   struct argv_entry * * end_of_done = &done_list;

   toparg->next = NULL;
   todo_list = toparg;

   /* Now loop through entries until nothing remains on the worklist
    */
   while ((av = todo_list) != NULL) {
      int done, lastslash, i, c, seenwild, nextslash;

      todo_list = av->next;
      av->next = NULL;

#ifdef DEBUGGING
      fprintf(stderr, "ExpandOneArg processing %s\n", av->name);
#endif

      i = av->scanpos;
      done = 0;
      lastslash = -1;
      nextslash = -1;
      seenwild = 0;
      while (! done) {
         switch (c = av->name[i]) {
         case '\0':
            if (seenwild) {
               nextslash = i;
            }
            done = 1;
            break;
         case ':':
            if ((av->scanpos == 0) && (lastslash == -1)) {
               /* If this scan is starting at the beginning of the string
                * and I have not yet seen a directory separator, then
                * treat a ':' like a directory separator since it divides
                * the device name from the file name
                */
               if (seenwild) {
                  nextslash = i;
                  done = 1;
               } else {
                  lastslash = i;
               }
            }
            break;
         case '/':
            if (use_slash) {
               /* If we are supposed to accept / as a directory separator,
                * then convert this one to a \ and treat it as such.
                */
               av->name[i] = '\\';
               c = '\\';
               /* Fall through to the \ code */
            } else {
               break;
            }
         case '\\':
            if (seenwild) {
               nextslash = i;
               done = 1;
            } else {
               lastslash = i;
            }
            break;
         case START_RANGE_CHAR:
         case '*':
         case '?':
            seenwild = 1;
            break;
         }
         ++i;
      }
      /* At this point I have scanned the name and found the next component
       * (if any) containing any wildcard characters. The interesting
       * variables are:
       *
       * seenwild - 0 if no wildcard, 1 if wildcard seen.
       * lastslash - index in name of last directory separator prior to
       *             the file component which contained the wildcard
       * nextslash - index in name of next directory separator following
       *             the component which contained the wildcard
       */
#ifdef DEBUGGING
      fprintf(stderr, "ExpandOneArg seenwild %s, lastslash %d, nextslash %d\n",
              seenwild ? "TRUE" : "FALSE", lastslash, nextslash);
#endif
      if (! seenwild) {
         /* No wildcard in name at all. If this was the original input
          * argument, then it doesn't have to actually exist, otherwise
          * this is an expanded name, and it shouldn't be on the list unless
          * it actually exists (names like \*\lib could otherwise wind up
          * on the list because the * expanded previously and this component
          * didn't have a wildcard in it, but it still needs checking).
          */
         if ((av == toparg) ||
             (GetFileAttributes(av->name) != 0xFFFFFFFF)) {
            *end_of_done = av;
            end_of_done = &av->next;
         }
      } else if ((lastslash == -1) && (av->name[nextslash] == ':')) {
         /* The wildcard pattern appeared prior to a ':' in the first
          * file component. In this case, I need to match the pattern
          * against each of the legal device letters rather than the
          * contents of any directory.
          */
         DWORD drivebits;
         char drvltr[2];
         char d;
         char * pattern;
         int tailen = strlen(&av->name[nextslash]);

         drivebits = GetLogicalDrives();
         drvltr[1] = '\0';
         pattern = (char *)NTCLImalloc(nextslash + 1);
         strncpy(pattern, av->name, nextslash);
         pattern[nextslash] = '\0';
         for (d = 'Z'; d >= 'A'; --d) {
            drvltr[0] = d;
            if (((drivebits & (((DWORD)1) << (d - 'A'))) != 0) &&
                NTCLIWildMatch(drvltr, pattern, case_sensitive)) {
               int tailen = strlen(&av->name[nextslash]);
               struct argv_entry * nav = (struct argv_entry *)
                  NTCLImalloc(sizeof(struct argv_entry) + tailen + 2);
               nav->next = NULL;
               nav->scanpos = 2;
               nav->name[0] = drvltr[0];
               strcpy(&nav->name[1], &av->name[nextslash]);
               nav->next = todo_list;
               todo_list = nav;
            }
         }
      } else {
         /* Read the directory in question, matching all the names against
          * the wildcard pattern, and pushing the resulting combined names
          * back on the todo list for additional processing. Free up the
          * original pattern when done expanding it into individual pieces.
          */
         char * dirname;
         char * pattern;
         int patlen;
         int tailen = strlen(&av->name[nextslash]);
         HANDLE dirhandle;
         WIN32_FIND_DATA wfd;
         struct argv_entry * end_of_list = todo_list;

         if (lastslash == -1) {
            /* The very first component was a wildcard - I need to
             * scan the directory using the pattern .\*.*
             */
            dirname = (char *)NTCLImalloc(6);
            strcpy(dirname, ".\\*.*");
         } else {
            int dirnamelen = lastslash + 5;
            dirname = (char *)NTCLImalloc(dirnamelen);
            strncpy(dirname, av->name, lastslash);
            strcpy(dirname + lastslash, "\\*.*");
         }
         patlen = nextslash - (lastslash + 1);
         pattern = (char *)NTCLImalloc(patlen + 1);
         strncpy(pattern, &av->name[lastslash + 1], patlen);
         pattern[patlen] = '\0';

#ifdef DEBUGGING
         fprintf(stderr, "ExpandOneArg calling FindFirstFile of %s\n", dirname);
#endif

         dirhandle = FindFirstFile(dirname, &wfd);
         if (dirhandle != INVALID_HANDLE_VALUE) {
            for ( ; ; ) {
               /* This test is kind of complicated, but it checks for
                * several things. It checks the hidden file attribute
                * (unless we are supposed to see hidden files) and it avoids
                * matching . or .. with a wildcard pattern (unless the
                * pattern actually starts with a .). Any files that make it
                * through all of that are checked against the pattern.
                */
               if ((see_hidden ||
                    ((wfd.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == 0)) &&
                   ((pattern[0] == '.') ||
                    ((strcmp(wfd.cFileName, ".") != 0) &&
                     (strcmp(wfd.cFileName, "..") != 0))) &&
                   NTCLIWildMatch(wfd.cFileName, pattern, case_sensitive)) {
                  int fnlen = strlen(wfd.cFileName);
                  int totlen = (lastslash + 1) + fnlen + tailen;
                  struct argv_entry * nav = (struct argv_entry *)
                     NTCLImalloc(sizeof(struct argv_entry) + totlen);
                  struct argv_entry ** cf;

                  nav->next = NULL;
                  nav->scanpos = lastslash + 1 + fnlen;
                  strncpy(nav->name, av->name, lastslash + 1);
                  strncpy(&nav->name[lastslash + 1], wfd.cFileName, fnlen);
                  strcpy(&nav->name[lastslash + 1 + fnlen],
                         &av->name[nextslash]);
                  /* Insertion sort the name onto the todo list, making sure
                   * to insert it somewhere prior to end_of_list (which is
                   * holding the todo_list entry prior to starting to read
                   * this directory).  Sort in "reverse" order, because todo
                   * entries get popped off and pushed on the end of the
                   * done list, which effectively reverses the todo list
                   * order.
                   */

#ifdef DEBUGGING
                  fprintf(stderr, "ExpandOneArg adding name %s\n", nav->name);
#endif

                  for (cf = &todo_list;
                       (*cf != end_of_list) &&
                       (strcmp(&nav->name[lastslash + 1],
                               &((*cf)->name[lastslash + 1])) > 0);
                       cf = &((*cf)->next)) ;
#ifdef DEBUGGING
                  fprintf(stderr, "ExpandOneArg found insert point.\n");
#endif
                  nav->next = *cf;
                  *cf = nav;
               }
               if (! FindNextFile(dirhandle, &wfd)) {
                  break;
               }
            }
            FindClose(dirhandle);
#ifdef DEBUGGING
            fprintf(stderr, "ExpandOneArg done with directory scan\n");
#endif
         }

         NTCLIfree((void *)dirname);
         NTCLIfree((void *)pattern);
         if ((done_list != NULL) || (av != toparg)) {
            NTCLIfree((void *)av);
         }
      }
   }
   if (done_list != NULL) {
      struct argv_expansion * rval;
      rval = (struct argv_expansion *)
         NTCLImalloc((sizeof(struct argv_expansion)));
      rval->head = done_list;
      rval->tail = end_of_done;
      return rval;
   } else {
      return NULL;
   }
}

/* Utility routine for building registry key strings and opening them for
// reading.
*/
static int
OpenVaKey(HKEY root, PHKEY newkeyp, ...)
{
   va_list ap;
   DWORD size = 0;
   char * keyvec;
   char * d;
   int rval;

   va_start(ap, newkeyp);
   for ( ; ; ) {
      char * arg = va_arg(ap, char *);
      if (arg == NULL) {
         break;
      }
      size += 1 + strlen(arg);
   }
   va_end(ap);

   keyvec = (char *)NTCLImalloc(size + 1);
   d = keyvec;

   va_start(ap, newkeyp);
   for ( ; ; ) {
      char * arg = va_arg(ap, char *);
      if (arg == NULL) {
         break;
      }
      if (d != keyvec) {
         *d++ = '\\';
      }
      strcpy(d, arg);
      d += strlen(d);
   }
   va_end(ap);
   if (RegOpenKeyEx(root,
                    keyvec,
                    0,
                    KEY_READ,
                    newkeyp) == ERROR_SUCCESS) {
      rval = 1;
   } else {
      rval = 0;
   }
   NTCLIfree((void *)keyvec);
   return rval;
}

/* Utility routine to read a key value. Mallocs the space to hold the entire
// value, and returns it in keyval with the length in keylen and the type of
// key in keytype.
*/
static int
ReadKeyval(
   HKEY khand,
   char * valname,
   DWORD * keylen,
   void ** keyval,
   DWORD * keytype)
{
   unsigned char dummy[1];
   *keylen = 0;
   if (RegQueryValueEx(khand,
                       valname,
                       NULL,
                       keytype,
                       dummy,
                       keylen) == ERROR_MORE_DATA) {
      *keyval = (void *)NTCLImalloc(*keylen);
      if (RegQueryValueEx(khand,
                          valname,
                          NULL,
                          keytype,
                          (unsigned char *)*keyval,
                          keylen) == ERROR_SUCCESS) {
         return 1;
      } else {
         NTCLIfree(*keyval);
      }
   }
   *keylen = 0;
   *keyval = NULL;
   return 0;
}

/* GetFlag looks up the value for a given boolen flag in the registry
// (expecting a DWORD value holding TRUE or FALSE). It looks first in the
// progname specific registry, then in "Global Defaults", then falls back on
// the default_value passed in as an arg.
*/
static int
GetFlag(NTCLIOptinfo oi, char * flagname, int default_value)
{
   HKEY keyhandle;
   void * val;
   DWORD len;

   if (OpenVaKey(HKEY_CURRENT_USER, &keyhandle,
          "Software", "NTCLI", oi->progname, NTCLI_VERSION, NULL)) {
      DWORD keytype;
      if (ReadKeyval(keyhandle, flagname, &len, &val, &keytype)) {
         if (len == sizeof(DWORD)) {
            int rval = ((*((DWORD *)val)) != 0);
            NTCLIfree(val);
            return rval;
         } else {
            NTCLIfree(val);
         }
      }
   }
   if (OpenVaKey(HKEY_CURRENT_USER, &keyhandle,
          "Software", "NTCLI", "Global Defaults", NTCLI_VERSION, NULL)) {
      DWORD keytype;
      if (ReadKeyval(keyhandle, flagname, &len, &val, &keytype)) {
         if (len == sizeof(DWORD)) {
            int rval = ((*((DWORD *)val)) != 0);
            NTCLIfree(val);
            return rval;
         } else {
            NTCLIfree(val);
         }
      }
   }
   return default_value;
}
