
/* Copyright (C) Tom Horsley, 1998. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "ntcli.h"

static char * progname = "rm";

static int recursive = 0;
static int never_prompt = 0;
static int always_prompt = 0;
static int can_prompt = 0;
static int is_interactive = 0;

static int RmOneFile(const char * fn);

static int
ReadPrompt(const char * fn, DWORD attr)
{
   char buf[80];
   char * bufp;
   char * sep;

   if (can_prompt) {
      fputs("rm ", stderr);
      fputs(fn, stderr);
      sep = " (";
      if (attr & FILE_ATTRIBUTE_HIDDEN) {
         fputs(sep, stderr);
         fputs("hidden", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_SYSTEM) {
         fputs(sep, stderr);
         fputs("system", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_READONLY) {
         fputs(sep, stderr);
         fputs("readonly", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_ARCHIVE) {
         fputs(sep, stderr);
         fputs("archive", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_COMPRESSED) {
         fputs(sep, stderr);
         fputs("compressed", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_DIRECTORY) {
         fputs(sep, stderr);
         fputs("directory", stderr);
         sep = ", ";
      }
      if (attr & FILE_ATTRIBUTE_NORMAL) {
         fputs(sep, stderr);
         fputs("normal", stderr);
         sep = ", ";
      }
      fputs(") y/n? ", stderr);
      bufp = fgets(buf, sizeof(buf), stdin);
      if ((bufp != NULL) && ((bufp[0] == 'y') || (bufp[0] == 'Y'))) {
         return 1;
      }
   }
   return 0;
}

typedef struct save_name_struct {
   struct save_name_struct * next;
   char name[1];
} save_name;

static save_name *
SaveName(const char * sn, save_name * next)
{
   if ((sn[0] == '.') &&
       ((sn[1] == '\0') || ((sn[1] == '.') && (sn[2] == '\0')))) {
      /* Ignore '.' and '..'
      */
      return next;
   } else {
      int len = strlen(sn);
      save_name * snp = NTCLImalloc(sizeof(save_name)+len);
      snp->next = next;
      strcpy(snp->name, sn);
      return snp;
   }
}

static void
FreeSaveList(save_name * lp)
{
   save_name * np;
   while (lp != NULL) {
      np = lp->next;
      NTCLIfree(lp);
      lp = np;
   }
}

static int
RmOneDir(const char * fn)
{
   int len;
   char * pat;
   HANDLE dirhand;
   WIN32_FIND_DATA fd;
   save_name * list = NULL;
   int errcount = 0;

   if (! recursive) {
      if (! never_prompt) {
         fprintf(stderr, "%s: %s is a directory, not removed.\n", progname, fn);
      }
      return 1;
   }
   /* traverse directory, removing children, then directory itself.
   */
   len = strlen(fn);
   pat = NTCLImalloc(len + 5);
   strcpy(pat, fn);
   if (pat[len-1] != '\\') {
      pat[len++] = '\\';
   }
   strcpy(&pat[len], "*.*");
   dirhand = FindFirstFile(pat, &fd);
   if (dirhand == INVALID_HANDLE_VALUE) {
      if (! never_prompt) {
         fprintf(stderr, "%s: Cannot read directory: %s\n", progname, fn);
      }
      NTCLIfree(pat);
      return 1;
   }
   list = SaveName(fd.cFileName, list);
   while (FindNextFile(dirhand, &fd)) {
      list = SaveName(fd.cFileName, list);
   }
   if (GetLastError() != ERROR_NO_MORE_FILES) {
      if (! never_prompt) {
         fprintf(stderr, "%s: Error reading directory: %s\n", progname, fn);
      }
      FreeSaveList(list);
      NTCLIfree(pat);
      return 1;
   }
   FindClose(dirhand);
   NTCLIfree(pat);
   while (list != NULL) {
      save_name * np;
      int nlen = strlen(list->name);
      char * fullname = NTCLImalloc(len + nlen + 1);
      strcpy(fullname, fn);
      fullname[len-1] = '\\';
      strcpy(&fullname[len], list->name);
      errcount += RmOneFile(fullname);
      NTCLIfree(fullname);
      np = list->next;
      NTCLIfree(list);
      list = np;
   }
   if (! RemoveDirectory(fn)) {
      if (! never_prompt) {
         fprintf(stderr, "%s: Unable to remove directory: %s\n", progname, fn);
      }
      ++errcount;
   }
   return errcount;
}

static int
RmOneFile(const char * fn)
{
   int yes = 1;
   DWORD attr = GetFileAttributes(fn);
   if (attr == 0xFFFFFFFF) {
      if (! never_prompt) {
         fprintf(stderr, "%s: no such file: %s\n", progname, fn);
      }
      return 1;
   } else {
      if (always_prompt ||
          ((! never_prompt) &&
           (attr & (FILE_ATTRIBUTE_HIDDEN |
                    FILE_ATTRIBUTE_READONLY |
                    FILE_ATTRIBUTE_SYSTEM)))) {
         /* Try to prompt for y/n action on the delete.
         */
         yes = ReadPrompt(fn, attr);
      }
      if (yes) {
         if (attr & FILE_ATTRIBUTE_READONLY) {
            /* Have to remove the readonly attribute before I can delete
            */
            SetFileAttributes(fn, attr & (~ FILE_ATTRIBUTE_READONLY));
         }
         if (attr & FILE_ATTRIBUTE_DIRECTORY) {
            return RmOneDir(fn);
         }
         if (! DeleteFile(fn)) {
            if (! never_prompt) {
               fprintf(stderr, "%s: file not deleted: %s\n", progname, fn);
            }
            return 1;
         }
      }
   }
   return 0;
}

int
main(int argc, char ** argv)
{
   NTCLIOptinfo oi;
   char * onearg;
   int errcount = 0;
   int argcount = 0;

   /* I'll be trying to wildcard expand things which might be on removable
   // devices, so tell the system to take a hike when it comes to displaying
   // error pop-up window's and just return an error code from the system
   // service instead.
   */
   SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOOPENFILEERRORBOX);

   oi = NTCLIInitOptinfo(argc, argv, "rfiI");
   progname = NTCLIGetProgname(oi);
   for ( ; ; ) {
      int opt = NTCLIGetNextOption(oi);
      if (opt == -1) {
         char * msg = NTCLIErrorString(oi);
         if (msg != NULL) {
            fprintf(stderr, "%s\n", msg);
            exit(2);
         } else {
            break;
         }
      } else {
         switch(opt) {
         case 'r':
            recursive = 1;
            break;
         case 'f':
            never_prompt = 1;
            break;
         case 'i':
            always_prompt = 1;
            break;
         case 'I':
            is_interactive = 1;
            break;
         }
      }
   }

   /* Let's see if we can figure out how to prompt (apparently this doesn't
   ** work for shell mode in emacs buffers, but I can't figure out how the
   ** devil to do a test which will work correctly, hence the -I option
   ** that can be used to force it to believe stdin is an interactive
   ** device)...
   */
   if (always_prompt || (! never_prompt)) {
      struct _stat sb;
      if (is_interactive ||
          ((_fstat(fileno(stdin), &sb) == 0) && (sb.st_mode & _S_IFCHR))) {
         can_prompt = 1;
      } else {
         can_prompt = 0;
      }
   }

   /* Now expand the arguments and process the files.
    */
   NTCLIExpandWildcards(oi);
   while ((onearg = NTCLIGetNextPositionalArg(oi)) != NULL) {
      errcount += RmOneFile(onearg);
      ++argcount;
   }
   if (argcount == 0) {
      if (! never_prompt) {
         fprintf(stderr, "%s: Missing argument(s)\n", progname);
      }
      ++errcount;
   }
   return (((errcount > 0) && (! never_prompt)) ? 2 : 0);
}
